/**
 * Copyright (C) 2010  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.ow2.bonita.env.EnvConstants;
import org.ow2.bonita.facade.privilege.Rule.RuleType;
import org.ow2.bonita.facade.privilege.impl.RuleImpl;
import org.ow2.bonita.facade.privilege.impl.SimpleRuleImpl;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Anthony Birembaut, Matthieu Chaffotte
 */
public final class DbMigration {

  private static final Logger LOG = LoggerFactory.getLogger(DbMigration.class);

  private DbMigration() {
  }

  /**
   * Migrate DB schema of the database defined in the environment.
   */
  public static void migrateDb(final String domain, final String configurationName, final String db) throws Exception {
    BonitaConstants.getBonitaHomeFolder();
    final SessionFactory sessionFactory = DbTool.getSessionFactory(domain, configurationName.replaceAll("-configuration", "-session-factory"));
    final String database = db.toLowerCase();
    LOG.info("Running " + database + " " + configurationName + " DB migration...");
    InputStream inputStream = null;
    try {
      inputStream = findMigrationScript(database);
      executeScript(sessionFactory, inputStream, database);
      LOG.info("---------------Schema migrated---------------");
      migrateCustomRules(sessionFactory);
      sessionFactory.close();
      LOG.info("---------------Data migrated---------------");
    } finally {
      if (inputStream != null) {
        inputStream.close();
      }
    }
  }

  public static void executeScript(SessionFactory sessionFactory, InputStream inputStream, String db) {
    final byte[] bytes = IoUtil.readBytes(inputStream);
    final String scriptContent = new String(bytes);
    final List<String> commands = getCommands(scriptContent, db);

    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();

    LOG.info("DB Commands Execution");
    for (String command : commands) {
      LOG.info("Executing command : " + command);
      try {
        session.createSQLQuery(command).executeUpdate();
      } catch (Exception e) {
        System.err.println("Error while executing command: " + command);
        LOG.error(e.getMessage(), e.getCause());
      }
    }
    session.getTransaction().commit();
    session.close();
  }

  public static List<String> getCommands(final String scriptContent, final String db) {
    String delimiter = ";";
    if (db.equals("sqlserver") || db.equals("sybase")) {
      delimiter = "go";
    }
    String regex = delimiter.concat("\n");
    final List<String> commands = new ArrayList<String>();
    String[] tmp = scriptContent.split(regex);
    for (String command : tmp) {
      commands.add(command.trim());
    }
    int lastIndex = commands.size() - 1;
    String lastCommand = commands.get(lastIndex);
    int index = lastCommand.lastIndexOf(delimiter);
    if (index > 0) {
      lastCommand = lastCommand.substring(0, index);
      commands.remove(lastIndex);
      commands.add(lastCommand);
    }
    return commands;
  }

  public static InputStream findMigrationScript(final String db) {
    final StringBuilder migrationScript = new StringBuilder();
    migrationScript.append("/migration/").append(db).append(".sql");
    LOG.info("Loading Script " + migrationScript.toString());
    final InputStream inputStream = DbMigration.class.getResourceAsStream(migrationScript.toString());
    if (inputStream == null) {
      final String message = ExceptionManager.getInstance().getFullMessage("bh_DBM_2");
      throw new IllegalArgumentException(message);
    }
    return inputStream;
  }

  public static void main(final String[] args) throws Exception {
    if (args == null || args.length != 2) {
      final String message = ExceptionManager.getInstance().getFullMessage("bh_DBM_1");
      throw new IllegalArgumentException(message);
    }
    final String domain = args[0];
    final String db = args[1].toLowerCase();
    String disableSearch = "bonita.search.use";
    System.setProperty(disableSearch, "false");
    LOG.info("Starting History DB migration");
    migrateDb(domain, EnvConstants.HB_CONFIG_HISTORY, db);
    LOG.info("Starting Core DB migration");
    migrateDb(domain, EnvConstants.HB_CONFIG_CORE, db);
    System.clearProperty(disableSearch);
  }

  @SuppressWarnings("unchecked")
  private static void migrateCustomRules(final SessionFactory sessionFactory) throws Exception {
    LOG.info("---------------Migrating rules---------------");
    Session session = null;
    try {
      session = sessionFactory.openSession();
      final Transaction tx = session.beginTransaction();

      final Query rulesStmt = session.createQuery("SELECT rule FROM org.ow2.bonita.facade.privilege.impl.RuleImpl AS rule WHERE rule.type = '" + RuleType.CUSTOM.name() + "'");
      final List<RuleImpl> rules = (List<RuleImpl>) rulesStmt.list();
      final Set<RuleImpl> ruleToDelete = new HashSet<RuleImpl>();
      RuleImpl newrule;
      for (RuleImpl rule : rules) {
        if (rule.getItems().contains("logout")) {
          newrule = new SimpleRuleImpl(RuleType.LOGOUT);
          newrule.setDescription(rule.getDescription());
          newrule.setLabel(rule.getLabel());
          newrule.setName(rule.getName() + "(1)");
          newrule.addEntities(rule.getEntities());
          newrule.addGroups(rule.getGroups());
          newrule.addMemberships(rule.getMemberships());
          newrule.addRoles(rule.getRoles());
          newrule.addUsers(rule.getUsers());
          session.save(newrule);
          ruleToDelete.add(rule);
        }
        if (rule.getItems().contains("change_password")) {
          newrule = new SimpleRuleImpl(RuleType.PASSWORD_UPDATE);
          newrule.setDescription(rule.getDescription());
          newrule.setLabel(rule.getLabel());
          newrule.setName(rule.getName() + "(2)");
          newrule.addEntities(rule.getEntities());
          newrule.addGroups(rule.getGroups());
          newrule.addMemberships(rule.getMemberships());
          newrule.addRoles(rule.getRoles());
          newrule.addUsers(rule.getUsers());
          session.save(newrule);
          ruleToDelete.add(rule);
        }

      }
      // delete old rules that have been migrated.
      for (RuleImpl rule : ruleToDelete) {
        session.delete(rule);
      }
      tx.commit();
    } catch (HibernateException e) {
      LOG.warn("Unable to migrate rules", e);
    } finally {
      if (session != null) {
        session.close();
      }
      LOG.info("---------------Rules Migrated---------------");
    }
  }
}
